(* 쥸 *)

open TreeAsm

let seq exp e = concat (Ret exp) reg_dummy e

(* for register coalescing *)
(* [XXX] Callä顢̵̣Ȥո̤ʤΤɤʤ
         ΤˡCalläɤפ֤ͤ1Ǥ˴ޤ롣 *)
let rec target' src dest = function
| Mov x when x = src && is_reg dest -> false, [dest]
| If(_,_,_,e1,e2) | IfI(_,_,_,e1,e2) | FIf(_,_,_,e1,e2) ->
    let c1, p1 = target src dest e1 in
    let c2, p2 = target src dest e2 in
    (c1 || c2, p1 @ p2)
| Call(_,ys) -> true, target_args src 0 ys
| _ -> (false, [])
and target src dest = function (* register targeting *)
| Ret exp -> target' src dest exp
| Let(x,exp,e) ->
    let c1, p1 = target' src x exp in
    if c1 then true, p1 else
    let c2, p2 = target src dest e in
    (c2, p1 @ p2)
and target_args src n = function (* auxiliary function for Call *)
| [] -> []
| y :: ys when src = y -> regs.(n) :: target_args src (n+1) ys
| _ :: ys -> target_args src (n+1) ys

let add x r regenv =
  if is_reg x then (assert (x = r); regenv) else
  M2.add x r regenv

exception NoReg of string
let find x regenv =
  if is_reg x then x else
  try M2.find x regenv
  with Not_found -> raise (NoReg x)

let rec alloc dest cont regenv x =
  (* allocate a register or fail *)
  assert (not (M2.mem x regenv));
  if is_reg x then x else
  let free = fv cont in
  try
    let _, prefer = target x dest cont in
    let live = (* Ƥ쥸 *)
      S2.fold
        (fun y live ->
          try S2.add (find y regenv) live
          with _ -> live)
        free S2.empty
    in
    List.find (* Ǥʤ쥸õ *)
      (fun r -> not (S2.mem r live))
      (prefer @ regs.(0) :: List.rev regs_lst)
  with Not_found ->
    failwith ("register allocation failed for " ^ x)

let rec g dest cont regenv = function (* ̿Υ쥸 *)
  | Ret exp -> g'_and_restore dest cont regenv exp
  | Let(x,exp,e) ->
      assert (not (M2.mem x regenv));
      let cont' = concat e dest cont in
      let (e1', regenv1) = g'_and_restore x cont' regenv exp in
      let r = alloc dest cont' regenv1 x in
      let (e2', regenv2) = g dest cont (add x r regenv1) e in
      (concat e1' r e2', regenv2)
and g'_and_restore dest cont regenv exp =
  (* Ѥѿ򥹥å쥸Restore *)
  try g' dest cont regenv exp
  with NoReg x -> g dest cont regenv (Let(x, Restore x, Ret exp))
and g' dest cont regenv exp =
  (* ̿Υ쥸 *)
  match exp with
  | Nop | Rd | Imm _ | Restore _ -> (Ret exp, regenv)
  | Mov x -> (Ret(Mov(find x regenv)), regenv)
  | Op2(o,x,y) -> Ret(Op2(o, find x regenv, find y regenv)), regenv
  | Op1i(o,x,i) -> Ret(Op1i(o, find x regenv, i)), regenv
  | Op1(o,x) -> Ret(Op1(o, find x regenv)), regenv
  | St(x,y,i) -> Ret(St(find x regenv, find y regenv, i)), regenv
  | If(c,x,y,e1,e2) -> g'_if dest cont regenv exp e1 e2
        (fun e1' e2' -> If(c, find x regenv, find y regenv, e1', e2'))
  | IfI(c,x,i,e1,e2) -> g'_if dest cont regenv exp e1 e2
        (fun e1' e2' -> IfI(c, find x regenv, i, e1', e2'))
  | FIf(c,x,y,e1,e2) -> g'_if dest cont regenv exp e1 e2
        (fun e1' e2' -> FIf(c, find x regenv, find y regenv, e1', e2'))
  | Call(l, ys) -> g'_call dest cont regenv exp ys (fun ys -> Call(l, ys))
  | Save(x, y) -> assert false
and g'_if dest cont regenv exp e1 e2 k =
  (* ifΥ쥸 *)
  let (e1', regenv1) = g dest cont regenv e1 in
  let (e2', regenv2) = g dest cont regenv e2 in
  let free = fv cont in
  let regenv' = (* ξ˶̤Υ쥸ѿ *)
    S2.fold
      (fun x regenv' ->
        try
	  if is_reg x then regenv' else
          let r1 = M2.find x regenv1 in
          let r2 = M2.find x regenv2 in
          if r1 <> r2 then regenv' else
	  M2.add x r1 regenv'
        with Not_found -> regenv')
      free M2.empty
  in
  (S2.fold
     (fun x e -> (* Ǥʤѿʬľ˥ *)
       if x = dest || not (M2.mem x regenv) || M2.mem x regenv' then e else
       seq (Save(M2.find x regenv, x)) e)
     free (Ret(k e1' e2')),
   regenv')
and g'_call dest cont regenv exp ys k =
  (* ؿƤӽФΥ쥸 *)
  (S2.fold
     (fun x e ->
       if x = dest || not (M2.mem x regenv) then e else
       seq (Save(M2.find x regenv, x)) e)
     (fv cont)
     (Ret(k (List.map (fun y -> find y regenv) ys))),
   M2.empty)

let h { name = x; args = ys; body = e } =
  (* ؿΥ쥸 *)
  let (i, arg_regs_r, regenv) =
    List.fold_left
      (fun (i, arg_regs, regenv) y ->
        let r = if y = reg_dummy then reg_dummy else regs.(i) in
        (i + 1,
	 r :: arg_regs,
	 add y r regenv))
      (0, [], M2.empty)
      ys in
  let e', _ = g regs.(0) (Ret(Mov regs.(0))) regenv e in
  { name = x; args = List.rev arg_regs_r; body = e'; stksiz = 0 }

let f xs = (* ץΤΥ쥸 *)
  Format.eprintf "register allocation ...@.";
  List.map h xs
